/* eslint-disable no-loop-func */
/* eslint-disable @typescript-eslint/no-unused-vars */
/* eslint-disable prefer-const */
'use strict';

let client = null;

let cloudMode = '';

let serverIp = [];
let port = null;
let targetIp = '';
let targetPort = '';
let https = '';
let mqttURL = [];
let mqttPort = null;
let mqttForceLogin = false;
let sdkMode = '';
let projectId = null;

let remoteDeviceId = '';
let password = '';

let showLog = true;
let reportLog = true;

// 现场设备 ID，支持同时拉取多个设备的视频流
let fieldDeviceIds = [];
// 视频流 ID
let streamIds = [];
// 是否同时拉取音频流
let audioStream = false;
// 自定义消息
let customMessage = '';
// 是否通过可靠通道发送自定义消息
let orderedDataChannel = true;

let gatewayList = [];
let connectedDeviceIds = [];
let permissions = {};
let subscribedStreamIds = {};
let customMessageReceived = {};

// Demo 是否已获取过麦克风权限
let hasMicrophonePermission = false;
// 本地音频设备列表
let localAudioDeviceList = [];
// 选中的本地音频设备 ID
let selectedLocalAudioDeviceId = '';
// 本地音频流
let localAudioTrack = null;
// 本地音频是否已推流
let isLocalAudioTrackPublished = false;

const sdkVersionElement = document.querySelector('#sdk-version');
sdkVersionElement.innerText = TRROSDK.version;

const serverIpInputElement = document.querySelector('#server-ip-input');
if (serverIpInputElement) {
  serverIpInputElement.addEventListener('change', ({ target: { value } }) => {
    serverIp = value
      .split(',')
      .map((v) => v?.trim?.() ?? v)
      .filter((v) => !!v);
  });
}

const portInputElement = document.querySelector('#port-input');
if (portInputElement) {
  portInputElement.addEventListener('change', ({ target: { value } }) => {
    port = Number(value);
  });
}

const projectIdInputElement = document.querySelector('#project-id-input');
projectIdInputElement.addEventListener('change', ({ target: { value } }) => {
  projectId = cloudMode === 'private' ? Number(value) : value;
});

const remoteDeviceIdInputElement = document.querySelector('#remote-device-id-input');
remoteDeviceIdInputElement.addEventListener('change', ({ target: { value } }) => {
  remoteDeviceId = value;
});

const passwordInputElement = document.querySelector('#password-input');
passwordInputElement.addEventListener('change', ({ target: { value } }) => {
  password = value;
});

const mqttForceLoginInputElement = document.querySelector('#mqtt-force-login-input');
mqttForceLoginInputElement.checked = mqttForceLogin;
mqttForceLoginInputElement.addEventListener('change', ({ target: { checked } }) => {
  mqttForceLogin = checked;
});

const fieldDeviceIdInputElement = document.querySelector('#field-device-id-input');
fieldDeviceIdInputElement.value = fieldDeviceIds.join(',');
fieldDeviceIdInputElement.addEventListener('change', ({ target: { value } }) => {
  fieldDeviceIds = value
    .split(',')
    .map((v) => v.trim())
    .filter((v) => v?.length);
});

const currentConnectionStateTextElement = document.querySelector('#current-connection-state-text');

const currentPermissionTextElement = document.querySelector('#current-permission-text');

const streamIdInputElement = document.querySelector('#stream-id-input');
streamIdInputElement.addEventListener('change', ({ target: { value } }) => {
  streamIds = value
    .split(',')
    .map((v) => v.trim())
    .filter((v) => v?.length)
    .map(({ value }) => Number(value));
});

const audioStreamInputElement = document.querySelector('#audio-stream-input');
audioStreamInputElement.addEventListener('change', ({ target: { checked } }) => {
  audioStream = checked;
});

const remoteDeviceCustomMessageInputElement = document.querySelector('#remote-device-custom-message-input');
remoteDeviceCustomMessageInputElement.addEventListener('change', ({ target: { value } }) => {
  customMessage = value;
});

const sendMsgViaOrderedDCInputElement = document.querySelector('#ordered-dc-input');
sendMsgViaOrderedDCInputElement.addEventListener('change', ({ target: { checked } }) => {
  orderedDataChannel = checked;
});

const fieldDeviceCustomMessageTextElement = document.querySelector('#field-messages-wrapper');

const localAudioDeviceSelectElement = document.querySelector('#local-audio-device-select');
localAudioDeviceSelectElement.addEventListener('change', ({ target: { selectedOptions } }) => {
  for (const { value } of selectedOptions) {
    selectedLocalAudioDeviceId = value;
  }
});

const notificationElement = document.querySelector('#notification');
const notificationBtmElement = document.querySelector('#notification-btm');

const init = async () => {
  try {
    client = new TRROSDK({
      cloudMode,
      ...(cloudMode === 'private' && serverIp?.length ? { serverIp } : {}),
      ...(cloudMode === 'private' && port !== undefined ? { port } : {}),
      ...(cloudMode === 'private' && !!targetIp ? { targetIp } : {}),
      ...(cloudMode === 'private' && targetPort !== undefined ? { targetPort } : {}),
      ...(cloudMode === 'private' ? { https } : {}),
      mqttOptions: {
        ...(mqttURL?.length ? { url: mqttURL } : {}),
        ...(mqttPort && Number.isInteger(mqttPort) ? { port: mqttPort } : {}),
        ...(mqttForceLogin !== undefined && mqttForceLogin !== null ? { forceLogin: mqttForceLogin } : {}),
      },
      ...(!!sdkMode ? { sdkMode } : {}),
      projectId,
      remoteDeviceId,
      password,
      observer: {
        onMqttConnectionState: ({ state, message }) => {
          console.log('onMqttConnectionState', state, message);
        },
        onSignalingState: ({ roomId, state, data }) => {
          console.log('onSignalingState', roomId, state, data);
        },
        onPeerConnectionState: ({ deviceId, streamId, mediaType, code, state, message, data }) => {
          console.log('onPeerConnectionState', deviceId, streamId, mediaType, code, state, message, data);
        },
        onKick: ({ code, reason }) => {
          console.log('onKick', code, reason);
          notification({
            title: 'onKick',
            description: `code ${code} reason ${reason}`,
          });
        },
        onConnected: ({ deviceId }) => {
          console.log('onConnected', deviceId);
          notification({
            title: 'onConnected',
            description: `fieldDeviceId ${deviceId}`,
          });
        },
        onDisconnected: ({ deviceId }) => {
          console.log('onDisconnected', deviceId);
          notification({
            title: 'onDisconnected',
            description: `fieldDeviceId ${deviceId}`,
          });
        },
        onTrackPublishState: ({ state, deviceId, streamId, mediaType }) => {
          console.log('onTrackPublishState', state, deviceId, streamId, mediaType);
          notification({
            title: 'onTrackPublishState',
            description: `deviceId ${deviceId} streamId ${streamId} mediaType ${mediaType} state ${state}`,
          });
        },
        onControlData: ({ deviceId, data, ordered }) => {
          console.log('onControlData', deviceId, data, ordered);

          const deviceIndex = connectedDeviceIds.findIndex(
            (id) => deviceId === `${projectId}/${id}` || deviceId === id,
          );

          if (deviceIndex !== -1) {
            if (typeof data === 'string') {
              customMessageReceived[connectedDeviceIds[deviceIndex]] = data;
            } else if (data instanceof ArrayBuffer) {
              customMessageReceived[connectedDeviceIds[deviceIndex]] = `[ArrayBuffer (${data.byteLength})]`;
            }

            renderCustomMessagesReceived();
          }
        },
        onEvent: ({ type, data }) => {
          if (type === 'autoplay') {
            notification({
              title: `onEvent ${type}`,
              description: data.message,
            });

            if (data.code === 0 && data.mediaType === 'audio') {
              const audioContainer = document.getElementById(`remote-audio-${data.fieldDeviceId}`);
              const audioElem = audioContainer?.querySelector?.('audio');

              if (audioElem) {
                audioElem.controls = true;
              }
            }
          }
          if (type === 'localTrackUnpublished') {
            notification({
              title: `onEvent ${type}`,
              description: '',
            });

            isLocalAudioTrackPublished = false;
          }
        },
        onError: () => {},
      },
      loggerOption: {
        showLog,
        reportLog,
      },
    });

    const { code, message } = await client.init();

    if (code !== 0) {
      notification({
        title: '初始化失败',
        description: message,
      });
    } else {
      notification({
        title: '初始化成功',
        description: '',
      });
    }
  } catch (err) {
    notification({
      title: '初始化失败',
      description: err?.message,
    });
  }
};

const createVideoContainer = (deviceId, length, streamIds, withAudio) => {
  const wrapperElem = document.querySelector('#remote-media-wrapper');

  let deviceWrapperElem = document.getElementsByClassName(`remote-media-${deviceId} remote-media`)?.[0];

  if (!deviceWrapperElem) {
    deviceWrapperElem = document.createElement('div');
    deviceWrapperElem.className = `remote-media-${deviceId} remote-media`;

    const title = document.createElement('h4');
    title.innerText = deviceId;

    deviceWrapperElem.appendChild(title);
    wrapperElem.appendChild(deviceWrapperElem);
  }

  let deviceVideosElem = deviceWrapperElem.getElementsByClassName(`remote-videos-${deviceId} remote-videos`)?.[0];

  if (!deviceVideosElem) {
    deviceVideosElem = document.createElement('div');
    deviceVideosElem.className = `remote-videos-${deviceId} remote-videos`;

    deviceWrapperElem.appendChild(deviceVideosElem);
  }

  const oldStreamIds = [...deviceVideosElem.childNodes.values()].map((elem) => elem.id.at(-1));

  if (oldStreamIds.join(',') !== streamIds.join(',')) {
    const newVideos = [];

    for (const streamId of streamIds) {
      let videoWrapperElem = document.getElementById(`remote-video-${deviceId}-${streamId}`);

      if (!videoWrapperElem) {
        videoWrapperElem = document.createElement('div');
        videoWrapperElem.id = `remote-video-${deviceId}-${streamId}`;
        videoWrapperElem.className = 'remote-video';

        const videoInfoElem = document.createElement('p');
        videoInfoElem.className = 'info';
        videoInfoElem.innerText = `streamId: ${streamId}`;

        videoWrapperElem.append(videoInfoElem);
      } else {
        videoWrapperElem = videoWrapperElem.parentNode.removeChild(videoWrapperElem);
      }

      newVideos.push(videoWrapperElem);
    }

    while (deviceVideosElem.children.length) {
      deviceVideosElem.removeChild(deviceVideosElem.children[0]);
    }
    deviceVideosElem.append(...newVideos);
  }

  let deviceAudioElem = deviceWrapperElem.getElementsByClassName(`remote-audios-${deviceId} remote-audios`)?.[0];

  if (!deviceAudioElem) {
    deviceAudioElem = document.createElement('div');
    deviceAudioElem.className = `remote-audios-${deviceId} remote-audios`;

    deviceWrapperElem.appendChild(deviceAudioElem);
  }

  let audioWrapperElem = document.getElementById(`remote-audio-${deviceId}`);

  if (!audioWrapperElem && withAudio) {
    audioWrapperElem = document.createElement('div');
    audioWrapperElem.id = `remote-audio-${deviceId}`;
    audioWrapperElem.className = 'remote-audio';

    deviceAudioElem.appendChild(audioWrapperElem);
  } else if (audioWrapperElem && !withAudio) {
    audioWrapperElem.parentNode.removeChild(audioWrapperElem);
  }
};

// 连接设备
const connect = async (id) => {
  if (!client) {
    console.log('connect action error, client not found');
    notification({
      title: 'connect action error',
      description: 'SDK 实例不存在',
    });
    return { code: -1, message: 'SDK 实例不存在' };
  }

  if (!id?.length) {
    console.log('connect action error, empty fieldDeviceId');
    notification({
      title: 'connect action error',
      description: 'empty fieldDeviceId',
    });
    return { code: -1, message: 'empty fieldDeviceId' };
  }

  const { code, message } = await client.connect(id);
  console.log('TRROSDK connect', id, code, message);
  if (code === 0) {
    connectedDeviceIds.push(id);
    notification({
      title: 'connect action success',
      description: `${id}, code: ${code}, message: ${message}`,
    });
  } else {
    notification({
      title: 'connect action error',
      description: `${id}, code: ${code}, message: ${message}`,
    });
  }
  renderConnectionStates();
  return { code, message };
};

// 订阅
const subscribe = async (id, streams, audioStream) => {
  if (!client) {
    console.log('subscribe action error, client not found');
    notification({
      title: 'subscribe action error',
      description: 'SDK 实例不存在',
    });
    return { code: -1, message: 'SDK 实例不存在' };
  }

  if (!connectedDeviceIds.includes(id)) {
    console.log('subscribe action error, field device not connected', id);
    notification({
      title: 'subscribe action error',
      description: 'field device not connected',
    });
    return { code: -1, message: 'field device not connected' };
  }

  if (!streams.length) {
    console.log('subscribe action error, empty streams');
    notification({
      title: 'subscribe action error',
      description: 'empty streams',
    });
    return { code: -1, message: 'empty streams' };
  }

  createVideoContainer(id, streams.length, streams, audioStream);

  const { code, message } = await client.subscribe({
    fieldDeviceId: id,
    containers: streams.map((streamId) => ({
      streamId,
      mount: `#remote-video-${id}-${streamId}`,
      callback: ({ latency }) => {
        const videoWrapperElem = document.getElementById(`remote-video-${id}-${streamId}`);

        if (videoWrapperElem) {
          const infoElem = videoWrapperElem.getElementsByClassName('info')?.[0];

          if (infoElem) {
            infoElem.innerHTML = `streamId: ${streamId}, latency: ${latency}`;
          }
        }
      },
    })),
    ...(audioStream ? { audioContainer: `#remote-audio-${id}` } : {}),
  });
  console.log('TRROSDK subscribe', id, streams, audioStream, code, message);
  if (code === 0 || code === 4) {
    subscribedStreamIds[id] = streams;
    notification({
      title: 'subscribe action success',
      description: `${id}, videoStreams: ${streams}, audioStream: ${audioStream}`,
    });
  } else {
    notification({
      title: 'subscribe action error',
      description: `${id}, code: ${code}, message: ${message}`,
    });
  }

  return { code, message };
};

// 取消订阅
const unsubscribe = async (id, streams, audioStream) => {
  if (!client) {
    console.log('unsubscribe action error, client not found');
    notification({
      title: 'unsubscribe action error',
      description: 'SDK 实例不存在',
    });
    return { code: -1, message: 'SDK 实例不存在' };
  }

  if (!connectedDeviceIds.includes(id)) {
    console.log('unsubscribe action error, field device not connected', id);
    notification({
      title: 'unsubscribe action error',
      description: 'field device not connected',
    });
    return { code: -1, message: 'field device not connected' };
  }

  if (!streams.length) {
    console.log('unsubscribe action error, empty streams');
    notification({
      title: 'unsubscribe action error',
      description: 'empty streams',
    });
    return { code: -1, message: 'empty streams' };
  }

  const { code, message } = await client.unsubscribe({
    fieldDeviceId: id,
    streamIds: streams,
    audioStream,
  });
  console.log('TRROSDK unsubscribe', id, streams, audioStream, code, message);
  if (code === 0) {
    subscribedStreamIds[id] = (subscribedStreamIds[id] ?? []).filter((i) => !streams.includes(i));
    if (!subscribedStreamIds[id].length) {
      delete subscribedStreamIds[id];
    }
    notification({
      title: 'unsubscribe action success',
      description: `${id}, videoStreams: ${streams}, audioStream: ${audioStream}`,
    });
  } else {
    notification({
      title: 'unsubscribe action error',
      description: `${id}, code: ${code}, message: ${message}`,
    });
  }
};

// 断开连接
const disconnect = async (id) => {
  if (!client) {
    console.log('disconnect action error, client not found');
    notification({
      title: 'disconnect action error',
      description: 'SDK 实例不存在',
    });
    return { code: -1, message: 'SDK 实例不存在' };
  }

  if (!connectedDeviceIds.includes(id)) {
    console.log('disconnect action error, field device not connected');
    notification({
      title: 'disconnect action error',
      description: 'field device not connected',
    });
    return { code: -1, message: 'field device not connected' };
  }

  const { code, message } = await client.disconnect(id);
  console.log('TRROSDK disconnect', id, code, message);
  if (code === 0) {
    connectedDeviceIds.splice(connectedDeviceIds.indexOf(id), 1);
    notification({
      title: 'disconnect action success',
      description: id,
    });
  } else {
    if (code === 1) {
      connectedDeviceIds.splice(connectedDeviceIds.indexOf(id), 1);
    }
    notification({
      title: 'disconnect action error',
      description: `${id}, code: ${code}, message: ${message}`,
    });
  }
  renderConnectionStates();
  return { code, message };
};

// 发起连接（连接设备 + 订阅媒体流）
const startConnection = async () => {
  console.log('startConnection action start');

  if (!client) {
    console.log('startConnection action error, client not found');
    notification({
      title: 'startConnection action error',
      description: 'SDK 实例不存在',
    });
    return;
  }

  if (!fieldDeviceIds?.length) {
    console.log('startConnection action error, empty fieldDeviceIds');
    notification({
      title: 'startConnection action error',
      description: 'empty fieldDeviceIds',
    });
    return;
  }

  const connected = fieldDeviceIds.filter((id) => connectedDeviceIds.includes(id));
  console.log('startConnection action, already connected', connected);
  const needToDisconnect = connectedDeviceIds.filter((id) => !connected.includes(id));
  console.log('startConnection action, need to disconnect', needToDisconnect);
  const needToConnect = fieldDeviceIds.filter((id) => !connectedDeviceIds.includes(id));
  console.log('startConnection action, need to connect', needToConnect);

  for (const id of connected) {
    const subscribed = subscribedStreamIds[id] ?? [];
    console.log('startConnection action, already subscribed', subscribed);
    const needToUnsubscibed = subscribed.filter((i) => !streamIds.includes(i));
    console.log('startConnection action, need to unsubscribe', needToUnsubscibed);
    const needToSubscibed = [
      ...subscribed.filter((i) => streamIds.includes(i)),
      ...streamIds.filter((i) => !subscribed.includes(i)),
    ];
    console.log('startConnection action, need to subscribe', needToSubscibed);

    if (needToUnsubscibed) {
      await unsubscribe(id, needToUnsubscibed, audioStream);
    }

    await subscribe(id, needToSubscibed, audioStream);
  }

  for (const id of needToDisconnect) {
    await disconnect(id);
  }

  for (const id of needToConnect) {
    const connectCode = await connect(id);

    if (connectCode) {
      await subscribe(id, streamIds, audioStream);
    }
  }
};

// 取消连接（取消订阅所有媒体流 + 断开连接）
const endConnection = async () => {
  console.log('endConnection action start');

  if (!client) {
    console.log('endConnection action error, client not found');
    notification({
      title: 'endConnection action error',
      description: 'SDK 实例不存在',
    });
    return;
  }

  for (const id of connectedDeviceIds) {
    const subscribed = subscribedStreamIds[id] ?? [];

    await unsubscribe(id, subscribed, audioStream);

    await disconnect(id);
  }
};

// 获取本地音频设备列表
const getLocalAudioDeviceList = async () => {
  try {
    if (!hasMicrophonePermission) {
      // 尝试获取麦克风权限
      const stream = await navigator.mediaDevices.getUserMedia({
        audio: true,
        video: false,
      });
      stream.getTracks().forEach((track) => track.stop());
    }

    const devices = await navigator.mediaDevices.enumerateDevices();
    const deviceList = devices
      .filter(({ kind }) => kind === 'audioinput')
      .map(({ deviceId, label }) => ({ id: deviceId, label }));

    console.log('getAudioDeviceList action success', deviceList);

    notification({
      title: 'getAudioDeviceList action success',
      description: `found ${deviceList.length} audio devices`,
    });

    localAudioDeviceList = deviceList;

    if (localAudioDeviceSelectElement) {
      // Chrome 86+
      localAudioDeviceSelectElement.replaceChildren();

      for (const { id, label } of localAudioDeviceList) {
        const optionElem = document.createElement('option');
        optionElem.value = id;
        optionElem.innerText = label;

        localAudioDeviceSelectElement.appendChild(optionElem);
      }
    }

    selectedLocalAudioDeviceId = '';
  } catch (err) {
    console.log('getLocalAudioDeviceList action error', err.name, err.message);

    notification({
      title: 'getLocalAudioDeviceList action error',
      description: `${err.name}, ${err.message}`,
    });
  }
};

// 若本地音频流意外中断，需要业务测自行取消推流
const onLocalAudioTrackEnded = () => {
  if (isLocalAudioTrackPublished) {
    unpublishLocalAudioTrack();
  }
};

// 创建本地音频流
const createLocalAudioTrack = async () => {
  if (localAudioTrack) {
    console.log('createLocalAudioTrack action error, already created');
    notification({
      title: 'createLocalAudioTrack action error',
      description: 'already created',
    });
    return;
  }

  try {
    const stream = await navigator.mediaDevices.getUserMedia({
      audio: !!selectedLocalAudioDeviceId ? { deviceId: selectedLocalAudioDeviceId } : true,
      video: false,
    });
    const [track] = stream.getAudioTracks();
    track.addEventListener('ended', () => {
      onLocalAudioTrackEnded();
    });

    console.log('createLocalAudioTrack action success', track.label, track.id);
    notification({
      title: 'createLocalAudioTrack action success',
      description: track.label,
    });

    localAudioTrack = track;
    if (localAudioDeviceSelectElement) {
      localAudioDeviceSelectElement.disabled = true;
    }
  } catch (err) {
    console.log('createLocalAudioTrack action error', err.name, err.message);
    notification({
      title: 'createLocalAudioTrack action error',
      description: `${err.name} ${err.message}`,
    });
  }
};

// 销毁本地音频流
const destroyLocalAudioTrack = async () => {
  if (!localAudioTrack) {
    console.log('destroyLocalAudioTrack action error, track not created');
    notification({
      title: 'destroyLocalAudioTrack action error',
      description: 'track not created',
    });
    return;
  }

  localAudioTrack.stop();
  localAudioTrack = null;
  onLocalAudioTrackEnded();

  console.log('destroyLocalAudioTrack action success');
  notification({
    title: 'destroyLocalAudioTrack action success',
    description: '',
  });

  if (localAudioDeviceSelectElement) {
    localAudioDeviceSelectElement.disabled = false;
  }
};

// 本地音频推流
const publishLocalAudioTrack = async () => {
  if (!client) {
    console.log('publishLocalAudioTrack action error, client not found');
    notification({
      title: '本地音频推流失败',
      description: 'SDK 实例不存在',
    });
    return;
  }

  if (isLocalAudioTrackPublished) {
    console.log('publishLocalAudioTrack action error, already published');
    notification({
      title: '本地音频推流失败',
      description: '当前已推流',
    });
    return;
  }

  if (!localAudioTrack) {
    console.log('publishLocalAudioTrack action error, track not created');
    notification({
      title: '本地音频推流失败',
      description: '本地音频流尚未创建',
    });
    return;
  }

  try {
    const { code, message } = await client.publish({
      audioTrack: localAudioTrack,
    });

    if (code === 0) {
      notification({
        title: '本地音频推流成功',
        description: message,
      });
      isLocalAudioTrackPublished = true;
    } else {
      console.log('publishLocalAudioTrack action error', code, message);

      if (code === 4) {
        notification({
          title: '本地音频推流失败',
          description: '没有订阅视频流',
        });
      } else if (code === 8) {
        notification({
          title: '本地音频推流失败',
          description: '无权限',
        });
      } else {
        notification({
          title: '本地音频推流失败',
          description: `code: ${code}, message: ${message}`,
        });
      }
    }
  } catch (err) {
    console.log('publishLocalAudioTrack action error', err.name, err.message);
    notification({
      title: '本地音频推流失败',
      description: `${err.name} ${err.message}`,
    });
  }
};

// 取消本地音频推流
const unpublishLocalAudioTrack = async () => {
  if (!client) {
    console.log('unpublishLocalAudioTrack action error, client not found');
    notification({
      title: 'unpublishLocalAudioTrack action error',
      description: 'SDK 实例不存在',
    });
    return;
  }

  if (!isLocalAudioTrackPublished) {
    console.log('publishLocalAudioTrack action error, not published');
    notification({
      title: 'publishLocalAudioTrack action error',
      description: 'not published',
    });
    return;
  }

  try {
    const { code, message } = await client.unpublish();

    if (code === 0) {
      notification({
        title: 'unpublishLocalAudioTrack action success',
        description: message,
      });
      isLocalAudioTrackPublished = true;
    } else {
      console.log('unpublishLocalAudioTrack action error', code, message);
      notification({
        title: 'unpublishLocalAudioTrack action error',
        description: message,
      });
    }
  } catch (err) {
    console.log('unpublishLocalAudioTrack action error', err.name, err.message);
    notification({
      title: 'unpublishLocalAudioTrack action error',
      description: `${err.name} ${err.message}`,
    });
  }
};

// 音频静音 or 音频取消静音
const muteAudioTrack = async (deviceId, muted) => {
  if (!client) {
    console.log('muteAudioTrack action error, client not found');
    notification({
      title: 'muteAudioTrack action error',
      description: 'SDK 实例不存在',
    });
    return;
  }

  try {
    const { code, message } = await client.muteAudio({ deviceId, muted });

    if (code === 0) {
      notification({
        title: 'muteAudioTrack action success',
        description: message,
      });
      isLocalAudioTrackPublished = true;
    } else {
      console.log('muteAudioTrack action error', code, message);
      notification({
        title: 'muteAudioTrack action error',
        description: message,
      });
    }
  } catch (err) {
    console.log('muteAudioTrack action error', err.name, err.message);
    notification({
      title: 'muteAudioTrack action error',
      description: `${err.name} ${err.message}`,
    });
  }
};

// 拉取网关列表
const pullGatewayList = async () => {
  if (!client) {
    console.log('pullGatewayList action error, client not found');
  }

  const { code, data, message } = await client.pullGatewayList();

  if (code !== 0) {
    console.log('pullGatewayList action error', code, message);

    notification({
      title: 'pullGatewayList error',
      description: message,
    });

    return;
  }

  if (!Array.isArray(data)) {
    console.log('pullGatewayList action error, invalid data', data);
    notification({
      title: 'pullGatewayList error',
      description: 'invalid data',
    });

    return;
  }

  gatewayList = data.map((gateway) => ({
    ...gateway,
    fieldDeviceId: gateway.deviceId.slice(projectId.length + 1),
  }));

  renderGatewayListTable();
};

// 渲染网关设备表格
const renderGatewayListTable = () => {
  const tableElem = document.getElementById('gateway-table');
  let tBodyElem = tableElem.getElementsByTagName('tbody')[0];

  if (tBodyElem) {
    tableElem.removeChild(tBodyElem);
  }

  tBodyElem = document.createElement('tbody');
  tableElem.append(tBodyElem);

  const ids = [];

  for (const { fieldDeviceId, deviceName, deviceType, version, status, streamCount } of gatewayList) {
    const trElem = document.createElement('tr');

    ids.push(fieldDeviceId);

    const checkboxElem = document.createElement('td');
    const checkboxInputElem = document.createElement('input');
    checkboxInputElem.type = 'checkbox';
    checkboxInputElem.id = `gateway-table-check-${fieldDeviceId}`;
    checkboxInputElem.addEventListener('change', ({ target: { checked } }) => {
      const idx = fieldDeviceIds.indexOf(fieldDeviceId);

      if (checked && idx === -1) {
        fieldDeviceIds.push(fieldDeviceId);
      } else if (!checked && idx !== -1) {
        fieldDeviceIds.splice(idx, 1);
      }

      fieldDeviceIdInputElement.value = fieldDeviceIds.join(',');

      const maxStreamCount = gatewayList
        .filter(({ fieldDeviceId }) => fieldDeviceIds.includes(fieldDeviceId))
        .reduce((prev, { streamCount }) => (streamCount > prev ? streamCount : prev), 0);
      renderStreamIdInput(maxStreamCount);
    });
    checkboxElem.append(checkboxInputElem);
    trElem.append(checkboxElem);

    const deviceIdElem = document.createElement('td');
    deviceIdElem.innerText = fieldDeviceId;
    trElem.append(deviceIdElem);

    const deviceNameElem = document.createElement('td');
    deviceNameElem.innerText = deviceName;
    trElem.append(deviceNameElem);

    const deviceTypeElem = document.createElement('td');
    deviceTypeElem.innerText = deviceType;
    trElem.append(deviceTypeElem);

    const versionElem = document.createElement('td');
    versionElem.innerText = version;
    trElem.append(versionElem);

    const statusElem = document.createElement('td');
    statusElem.innerText = status;
    trElem.append(statusElem);

    const streamCountElem = document.createElement('td');
    streamCountElem.innerText = streamCount;
    trElem.append(streamCountElem);

    tBodyElem.append(trElem);
  }

  const checkAllElem = document.getElementById('gateway-table-check-all');

  if (ids.length) {
    checkAllElem.disabled = false;
    checkAllElem.onchange = ({ target: { checked } }) => {
      for (const id of ids) {
        const idx = fieldDeviceIds.indexOf(id);
        const checkboxInputElem = document.getElementById(`gateway-table-check-${id}`);

        if (checkboxInputElem) {
          checkboxInputElem.checked = checked;
        }

        if (checked && idx === -1) {
          fieldDeviceIds.push(id);
        } else if (!checked && idx !== -1) {
          fieldDeviceIds.splice(idx, 1);
        }
      }

      fieldDeviceIdInputElement.value = fieldDeviceIds.join(',');

      const maxStreamCount = gatewayList
        .filter(({ fieldDeviceId }) => fieldDeviceIds.includes(fieldDeviceId))
        .reduce((prev, { streamCount }) => (streamCount > prev ? streamCount : prev), 3);
      renderStreamIdInput(maxStreamCount);
    };
  } else {
    checkAllElem.disabled = true;
    checkAllElem.onchange = null;
  }
};

// 控制权限请求
const requestPermission = async (permission) => {
  if (!client) {
    console.log('requestPermission action error, client not found');
    notification({
      title: '请求权限失败',
      description: 'SDK 实例不存在',
    });
    return;
  }

  if (!connectedDeviceIds.length) {
    console.log('requestPermission action error, no field device connected');
    notification({
      title: '请求权限失败',
      description: '当前没有已连接的设备',
    });
    return;
  }

  for (const deviceId of connectedDeviceIds) {
    const { code, data, message } = await client.requestPermission(deviceId, {
      permission,
    });

    if (code === 0) {
      console.log('requestPermission action success', deviceId, data);
      permissions[deviceId] = data.permission;
      notification({
        title: '请求权限成功',
        description: `${deviceId}, permission: ${data.permission}`,
      });
    } else {
      console.log('requestPermission action error', deviceId, permission, code, message);
      notification({
        title: '请求权限失败',
        description: `${deviceId}, code: ${code}, message: ${message}`,
      });
    }
  }
  renderPermissions();
};

// 发送自定义消息
const sendCustomMessageViaDataChannel = async () => {
  if (!client) {
    console.log('sendCustomMessageViaDataChannel action error, client not found');
    notification({
      title: '消息发送失败',
      description: 'SDK 实例不存在',
    });
    return;
  }

  if (!connectedDeviceIds.length) {
    console.log('sendCustomMessageViaDataChannel action error, no field device connected');
    notification({
      title: '消息发送失败',
      description: '当前没有已连接的设备',
    });
    return;
  }

  for (const deviceId of connectedDeviceIds) {
    const { code, message } = await client.sendControlData(deviceId, {
      data: customMessage,
      ordered: orderedDataChannel,
    });

    if (code === 0) {
      console.log('sendCustomMessageViaDataChannel action success', deviceId);
      notification({
        title: '消息发送成功',
        description: `${deviceId}`,
      });
    } else {
      console.log('sendCustomMessageViaDataChannel action error', deviceId, code, message);

      if (code === 42) {
        notification({
          title: '消息发送失败',
          description: `并未取得 ${deviceId} 的 master 权限`,
        });
      } else if (code === 45) {
        notification({
          title: '消息发送失败',
          description: `尚未订阅 ${deviceId} 的任何视频流`,
        });
      } else {
        notification({
          title: '消息发送失败',
          description: `${deviceId}, code: ${code}, message: ${message}`,
        });
      }
    }
  }
};

const renderConnectionStates = () => {
  if (currentConnectionStateTextElement) {
    const list = [];
    for (const fieldDeviceId of fieldDeviceIds) {
      const connected = connectedDeviceIds.includes(fieldDeviceId);
      list.push(
        `<b>${fieldDeviceId}</b> <span class="${connected ? 'success-text' : 'error-text'}">${
          connected ? '已连接' : '未连接'
        }</span>`,
      );
    }
    currentConnectionStateTextElement.innerHTML = list.join('<br/>');
  }
};

const renderPermissions = () => {
  if (currentPermissionTextElement) {
    const list = [];
    for (const [fieldDeviceId, permission] of Object.entries(permissions)) {
      list.push(`<b>${fieldDeviceId}</b> <span>${!!permission ? permission : '-'}<span>`);
    }
    currentPermissionTextElement.innerHTML = list.join('<br/>');
  }
};

const renderStreamIdInput = (streamCount) => {
  if (!streamIds.length || streamCount) {
    streamIds = Array(streamCount ?? 3)
      .fill(0)
      .map((_, i) => i);
  }

  streamIdInputElement.value = streamIds.join(',');
};

const renderCustomMessagesReceived = () => {
  if (fieldDeviceCustomMessageTextElement) {
    const list = [];
    for (const [fieldDeviceId, message] of Object.entries(customMessageReceived)) {
      list.push(`<b>${fieldDeviceId}</b> ${!!message ? message : '-'}`);
    }
    fieldDeviceCustomMessageTextElement.innerHTML = list.join('<br/>');
  }
};

const notification = ({ title, description }) => {
  const text = `${title}${!!description ? ` | ${description}` : ''}`;
  notificationElement.innerHTML = text;
  notificationBtmElement.innerHTML = text;
};
